package FluxY_Lib::MS::DataClass_GY;

################################################################################
# This class provides support for glycerol compound fragments, derivatized 
# using either TMS or TFAA.
#
# This class inherits basic functionality from the DataClass parent class,
# initializing it with the glycerol compounds, the atom list and compound name.
# It extends the basic DataClass to add support for derivatized fragments. As
# such it over-rides the GetCarbonInfo() and GetAtomCount() methods, as well as
# implementing the required PrintMassTables(). Three additional helper classes,
# CalcTotalAtoms(), GenerateMassList() and GetMassList() are also implemented.
#
################################################################################

use strict;
use warnings;

use FluxY_Lib::Chemistry::atomInfo; # User Lib - Atomic Masses.
use FluxY_Lib::Math::Math_Utils;    # User Lib - Simple math functions.

use FluxY_Lib::MS::DataClass;       # User Lib - The common parent class.
use FluxY_Lib::MS::DAgent;          # User Lib - Provides Derivatizing Agent Info.

our @ISA = qw(FluxY_Lib::MS::DataClass);    # Inherits from DataClass


# List of the atoms found in glycerol and subsequently considerred for natural
# abundance.
# Note: F & Si are included for convenience (commonly found in derivatizing agents).
my @atomList = ('C','H','O','F','Si');
my @atomMass = ();

# Each of the following atom counts is an array of the number of each corresponding
# atom relative to the order specified in the atomList, i.e. C, H, 0, F and Si.

# Hash of compound data, (first two array elements are required).
# 1. The compound names (glycerol), in lower case (for standard comparison).
# 2. The atom count for the glycerol compound.
my %compounds = (
   glyc => ["glycerol", [3,8,3,0,0]],
   );

# A 2D Hash of compound masses, indexed by compunds key and frags keys.
# Each element is the truncated mass value for the given compound/fragment.
my %massList = ();

# Variables initialized by the class:
my $dAgentName = "";   # The name of the derivatizing agent.
my @dAgent     = ();   # The atom count for the derivatizing agent.

my %frags      = ();   # Derivatized glucose fragment list.

# Hash of fragment info for each TMS derivatised glycerol fragment:
# - array 1: atom counts, and
# - array 2: the first and last positions of experimentally labelled carbons.
my %tmsFrags = (
   m000  => [[12,32,3,0,3], [1,3]], # M-0  - original mass, added for convenience.
	m015  => [[11,29,3,0,3], [1,3]], # M-15 - loss of a CH3.
	m090  => [[9,22,2,0,2], [1,3]],  # M-90 - loss of a (CH3)3SiOH.
   );

# Hash of fragment info for each TFAA derivatised glycerol fragment:
# - array 1: atom counts, and
# - array 2: the first and last positions of experimentally labelled carbons.
my %tfaaFrags = (
   m000  => [[9,5,6,9,0], [1,3]],   # M - original mass, added for convenience.
   m113  => [[7,5,4,6,0], [1,3]],   # M-113.
   m127  => [[6,3,4,6,0], [1,2]],   # M-127.
   );


################################################################################
# Class constructor
#
# Used to instantiate an object for this class.
#
################################################################################
sub new {
   my ($class, $dAgent) = @_;

   # Initialize the atom mass vector.
   @atomMass = &GetElementMasses(@atomList);

   # If not specified, set dAgent to TMS by default, otherwise check that the
   # derivatizing agent is supported (TMS, TFAA) for glycerol, or terminate.
   if ( !(defined($dAgent)) ) {
      $dAgent = "TMS";
		%frags  = %tmsFrags;
   } elsif ( lc($dAgent) eq 'tms' ) {
		%frags = %tmsFrags;
   } elsif ( lc($dAgent) eq 'tfaa' ) {
		%frags = %tfaaFrags;
	} else {
      print "\n\n\tThe specified derivatizing agent, '" . $dAgent . "', could not be used.\n\n";
      print "\tCurrently only TMS and TFAA are supported for derivatized glycerol.\n\n";
   
      exit -1;
   }

   # Set the derivatizing agent info:
	my ($dName, $atomCount, $atomList) = &GetDAgent($dAgent);
	
	$dAgentName  = $dName;
	my @atomCountDA = @{$atomCount};
	my @atomListDA  = @{$atomList};

   # Ensure that the derivatizing agent is in the same order as the atomList.
   foreach my $atom (@atomList) {
      my $found = 0;
      for (my $i=0; $i<@atomListDA; $i++) {
         if ( $atom eq $atomListDA[$i] ) {
            $found = 1;
            push(@dAgent, $atomCountDA[$i]);
            last;
         }
      }
      if ( !$found ) { push(@dAgent, 0); }
   }
	
   # Initialize the compound mass list.
   GenerateMassList();

   # Call the constructor of the parent class, Person.
   my $self = $class->SUPER::new(\%compounds, \@atomList, "Glycerol (GY)");

   bless $self, $class;

   return $self;
}


########################################################
#                                                      #
#   Over-ridden DataClass methods.                     #
#                                                      #
#   - to support the derivating agent and fragments.   #
#                                                      #
########################################################

################################################################################
# Subroutine: GetCarbonInfo
#
# Returns the number of carbon atoms, and the relative positional carbon string
# for the given fragment.
#
# Passed value $compound is not used, but included for compatability.
#
################################################################################
sub GetCarbonInfo{
   my ($self, $compound, $fragMass) = @_;

   foreach my $frag (sort keys %frags) {
      if ($fragMass == $massList{$compound}{$frag}) {
         my ($firstC,$lastC) = @{${$frags{$frag}}[1]};
            
         my $numCarbons = $lastC - $firstC + 1;
            
         my $fragAtomStr = "";
         while ( $firstC < $lastC ) {
            $fragAtomStr .= "$firstC,";
            $firstC++;
         }         
         $fragAtomStr .= "$lastC";
         
         return ($numCarbons, $fragAtomStr);
      }
   }

   return;
}


################################################################################
# Subroutine: GetAtomCount
#
# Returns the atom count corresponding to the given derivatized Glycerol
# fragment mass.
#
# Passed value $compound is not used, but included for compatability.
#
################################################################################
sub GetAtomCount{
   my ($self, $compound, $fragMass) = @_;
   
   foreach my $frag (sort keys %frags) {      
      if ($fragMass == $massList{$compound}{$frag}) { return @{${$frags{$frag}}[0]}; }
   }

   return;
}


######################################
#                                    #
#   Implemented DataClass methods.   #
#                                    #
######################################

################################################################################
# Subroutine: PrintMassTables
#
# Prints a table containing the atom count / fragment mass for Glycerol.
#
################################################################################
sub PrintMassTables{
   my $outFile = "GY_Data_Table.txt";

   open (FILE, ">" . $outFile) or FileSystemError(900, $outFile, $!);
   
   print FILE "The derivatizing agent is: " . $dAgentName . "\n";
   
   foreach my $compound (sort keys %compounds) {
      print FILE "\nFor: " . $compound . "\n";
   
      print FILE "\t" . join("\t", @atomList) . "\tMass\tC-Backbone\n";
            
      foreach my $frag (sort keys %frags) {
         print FILE $frag . "\t" . join("\t", @{${$frags{$frag}}[0]}) . "\t" . $massList{$compound}{$frag} . "\t" . join("-", @{${$frags{$frag}}[1]}) . "\n";
      }
   }
   
   close(FILE)                 or FileSystemError(901, $outFile, $!);
}


#######################
#                     #
#   Public methods.   #
#                     #
#######################

################################################################################
# Subroutine: CalcTotalAtoms
#
# Calculates the total number of atoms for glucose, unfragmented and
# derivatized.
#
# - The  derivatizing agent (TMS, AFAA, ...) attaches to each of the
# three carbons (replacong one Hydrogen at each bond).
#
################################################################################
sub CalcTotalAtoms{
   my ($compound) = @_;

	my @totAtoms = @{${$compounds{$compound}}[1]};

   my $multiplier = 3;   # The Number of derivatizing agent molecules.
   
   for(my $i=0; $i<@totAtoms; $i++) {
      $totAtoms[$i] += ($dAgent[$i] * $multiplier);
      
      # Correct for Hydrogen loss, 1 per dAgent:
      if ( $atomList[$i] eq "H" ) { $totAtoms[$i] -= $multiplier; }
   }

   return @totAtoms;
}


################################################################################
# Subroutine: GenerateMassList
#
# Generates a list of mass values in a hash (indexed by compound key) of hashes
# (indexed by fragment key).
#
################################################################################
sub GenerateMassList {
   foreach my $compound (sort keys %compounds) {
   
      my %list = ();
      
      foreach my $frag (sort keys %frags) {
         my @fragAtomCount = @{${$frags{$frag}}[0]};

         my $mass = 0;
         for(my $i=0; $i<@fragAtomCount; $i++) {
            $mass += ($fragAtomCount[$i] * $atomMass[$i]);
         }
         
         $list{$frag} = &Trunc($mass);
      }
      
      $massList{$compound} = \%list;
   }
}


################################################################################
# Subroutine: GetMassList
#
# Return the massList hash.
#
################################################################################
sub GetMassList {
   return %massList;
}

1;